@@ -23,7 +23,6 @@ from kazoo.protocol import paths
 import six
 
 from tooz import coordination
-from tooz.drivers import _retry
 from tooz import locking
 from tooz import utils
 
@@ -35,31 +34,22 @@ class ZooKeeperLock(locking.Lock):
         self.acquired = False
 
     def acquire(self, blocking=True):
-
-        @_retry.retry(stop_max_delay=blocking)
-        def _lock():
-            # NOTE(sileht): kazoo lock looks broken:
-            # https://github.com/python-zk/kazoo/issues/291
-            # so we track the exclusivity internally
-            if self.acquired is True:
-                if blocking:
-                    raise _retry.Retry
-                return False
-
-            if self._lock.acquire(blocking=bool(blocking),
-                                  timeout=0):
-                self.acquired = True
-                return True
-
-            if blocking:
-                raise _retry.Retry
-            return False
-
-        return _lock()
+        if isinstance(blocking, bool):
+            timeout = None
+        else:
+            blocking = True
+            timeout = float(blocking)
+        self.acquired = self._lock.acquire(blocking=blocking,
+                                           timeout=timeout)
+        return self.acquired
 
     def release(self):
-        self._lock.release()
-        self.acquired = False
+        if self.acquired:
+            self._lock.release()
+            self.acquired = False
+            return True
+        else:
+            return False
 
 
 class BaseZooKeeperDriver(coordination.CoordinationDriver):
@@ -68,13 +58,16 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
     :param timeout: connection timeout to wait when first connecting to the
                     zookeeper server
     """
-
-    _TOOZ_NAMESPACE = b"tooz"
+    #: Default namespace when none is provided.
+    TOOZ_NAMESPACE = b"tooz"
 
     def __init__(self, member_id, parsed_url, options):
         super(BaseZooKeeperDriver, self).__init__()
+        options = utils.collapse(options)
+        self._options = options
         self._member_id = member_id
-        self.timeout = int(options.get('timeout', ['10'])[-1])
+        self.timeout = int(options.get('timeout', '10'))
+        self._namespace = options.get('namespace', self.TOOZ_NAMESPACE)
 
     def _start(self):
         try:
@@ -85,7 +78,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                                           cause=e)
 
         try:
-            self._coord.ensure_path(self.paths_join("/", self._TOOZ_NAMESPACE))
+            self._coord.ensure_path(self.paths_join("/", self._namespace))
         except exceptions.KazooException as e:
             coordination.raise_with_cause(coordination.ToozError,
                                           "operation error: %s" % (e),
@@ -226,7 +219,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             return set(m.encode('ascii') for m in members_ids)
 
     def get_members(self, group_id):
-        group_path = self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
+        group_path = self.paths_join("/", self._namespace, group_id)
         async_result = self._coord.get_children_async(group_path)
         return ZooAsyncResult(async_result, self._get_members_handler,
                               timeout_exception=self._timeout_exception,
@@ -301,16 +294,16 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             return set(g.encode('ascii') for g in group_ids)
 
     def get_groups(self):
-        tooz_namespace = self.paths_join("/", self._TOOZ_NAMESPACE)
+        tooz_namespace = self.paths_join("/", self._namespace)
         async_result = self._coord.get_children_async(tooz_namespace)
         return ZooAsyncResult(async_result, self._get_groups_handler,
                               timeout_exception=self._timeout_exception)
 
     def _path_group(self, group_id):
-        return self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
+        return self.paths_join("/", self._namespace, group_id)
 
     def _path_member(self, group_id, member_id):
-        return self.paths_join("/", self._TOOZ_NAMESPACE,
+        return self.paths_join("/", self._namespace,
                                group_id, member_id)
 
     @staticmethod
@@ -338,7 +331,7 @@ class KazooDriver(BaseZooKeeperDriver):
 
     def __init__(self, member_id, parsed_url, options):
         super(KazooDriver, self).__init__(member_id, parsed_url, options)
-        self._coord = self._make_client(parsed_url, options)
+        self._coord = self._make_client(parsed_url, self._options)
         self._member_id = member_id
         self._timeout_exception = self._coord.handler.timeout_exception
 
@@ -464,23 +457,25 @@ class KazooDriver(BaseZooKeeperDriver):
         return ZooKeeperLock(
             name,
             self._coord.Lock(
-                self.paths_join(b"/", self._TOOZ_NAMESPACE, b"locks", name),
+                self.paths_join(b"/", self._namespace, b"locks", name),
                 self._member_id.decode('ascii')))
 
-    def run_watchers(self):
+    def run_watchers(self, timeout=None):
         ret = []
         while self._watchers:
             cb = self._watchers.popleft()
             ret.extend(cb())
-
         for group_id in six.iterkeys(self._hooks_elected_leader):
-            if self._get_group_leader_lock(group_id).acquire(blocking=False):
+            leader_lock = self._get_group_leader_lock(group_id)
+            if leader_lock.is_acquired:
+                # Previously acquired/still leader, leave it be...
+                continue
+            if leader_lock.acquire(blocking=False):
                 # We are now leader for this group
                 self._hooks_elected_leader[group_id].run(
                     coordination.LeaderElected(
                         group_id,
                         self._member_id))
-
         return ret
 
 
